# Copyright 2016 Pavle Jonoski
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = 'pavle'


class App:

    def __init__(self, name=None, description=None, command=None, params=None, needs=None):
        self.name = name
        self.description = description
        self.command = command
        self.params = params or {}
        self.needs = needs or { "cpu": 0, "memory": 0, "network": 0, "disk": 0 }
